import GPUtil
from tabulate import tabulate
import glog as log


class GPUUtility:
    def __init__(
        self,
        limit=1,
        priority="memory",
        minFreeLoad=0.0,
        minFreeMemory=0.0,
        minFreeMemoryMb=0.0,
        ignoreIDs=[],
        ignoreUUIDs=[],
    ):
        """Base class for any GPU-friendly objects. Simply wraps GPUtil module

        Args:
            limit (int): Max number of desired GPUs
            priority (str): How to prioritize which available GPUs to select
                Options:
                    "memory" (default): Selects qualifying GPUs with the most available memory
                    "first": Selects qualifying GPUs by ascending GPU id
                    "last": Selects qualifying GPUs by descending GPU id
                    "random": Selects qualifying GPUs randomly
                    "load": Selects qualifying GPUs with the most available load

            minFreeLoad (float): Percentage of available load needed by module
            minFreeMemory (float): Percentage of available memory needed by module
            minFreeMemoryMb (float): Megabytes of available memory needed by module
            ignoreIDs (list[int]): GPU to ignore by ID
            ignoreUUIDs (list[int]): GPUs to ignore by UUID
        """
        priority_options = set(["memory", "first", "last", "random", "load"])
        if not priority in priority_options:
            priority = "memory"

        self.limit = limit
        self.priority = priority
        self.minFreeLoad = minFreeLoad

        self.minFreeMemory = minFreeMemory
        self.minFreeMemoryMb = minFreeMemoryMb

        self.ignoreIDs = ignoreIDs
        self.ignoreUUIDs = ignoreUUIDs

    @property
    def GPUs(self):
        return self.get_gpus()

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, limit):
        self._limit = max(int(limit), 0)

    @property
    def minFreeMemory(self):
        return self._minFreeMemory

    @minFreeMemory.setter
    def minFreeMemory(self, mem):
        self._minFreeMemory = self._round_to_between_0_and_1(mem)

    @property
    def minFreeLoad(self):
        return self._minFreeLoad

    @minFreeLoad.setter
    def minFreeLoad(self, load):
        self._minFreeLoad = self._round_to_between_0_and_1(load)

    @staticmethod
    def _round_to_between_0_and_1(value):
        """Returns a value, element of [0,1]

        Args:
            value (float): Any number

        Returns:
            0 if value <= 0
            value if 0 < value < 1
            1 if value >= 1
        """
        value = max(value, 0)
        value = min(value, 1)
        return value

    def get_gpus(self, **kwargs):
        """Gets a list of qualifying GPU IDs
        """
        max_load = self._round_to_between_0_and_1(1.0 - self.minFreeLoad)
        max_mem = self._round_to_between_0_and_1(1.0 - self.minFreeMemory)

        log.debug("GPU Requirements")

        table = [
            ("order", self.priority),
            ("maxLoad", max_load),
            ("maxMemory", max_mem),
            ("excludeID", self.ignoreIDs),
            ("excludeUUID", self.ignoreUUIDs),
        ]

        table = tabulate(table, headers=["Parameter", "Value"], tablefmt="simple")
        log.debug(table)

        availableGPUids = GPUtil.getAvailable(
            order=self.priority,
            maxLoad=max_load,
            maxMemory=max_mem,
            excludeID=self.ignoreIDs,
            excludeUUID=self.ignoreUUIDs,
        )
        log.debug("GPU Util Found GPU IDs: " + str(availableGPUids))
        availableGPUids = self._filter_gpus(availableGPUids)
        log.debug("Filtered GPU IDs are: " + str(availableGPUids))

        return availableGPUids

    def _filter_gpus(self, availableGPUids):
        """Filters out GPUs by limit and available memory in megabytes

        Args:
            availableGPUids (list[int]): A list of candidate GPU IDs

        Returns:
            list[int]: A list qualifying of GPU IDs
        """
        GPUs = GPUtil.getGPUs()
        usableGPUids = []
        for GPUid in availableGPUids:
            if len(usableGPUids) == self.limit:
                break

            GPU = GPUs[GPUid]
            if GPU.memoryFree >= self.minFreeMemoryMb:
                usableGPUids.append(GPUid)

        return usableGPUids


class CupyUtility(GPUUtility):
    import numpy
    import importlib.util

    if importlib.util.find_spec("cupy"):
        import cupy
    else:
        cupy = None

    @property
    def xp(self):
        if len(self.GPUs) > 0 and not self.cupy is None:
            return self.cupy
        return self.numpy



import os
import sys
import glog as log
import numpy as np
import traceback
import importlib
import numbers

from multivitamin.module import ImagesModule
from multivitamin.data.response.utils import (
    p0p1_from_bbox_contour,
    crop_image_from_bbox_contour,
)
from multivitamin.module.utils import min_conf_filter_predictions
from multivitamin.data.response.dtypes import Region, Property

glog_level = os.environ.get("GLOG_minloglevel", None)

if glog_level is None:
    os.environ["GLOG_minloglevel"] = "1"
    log.info("GLOG_minloglevel isn't set. Setting level to 1 (info)")
    log.info(
        """GLOG_minloglevel levels are...
                    \t0 -- Debug
                    \t1 -- Info
                    \t2 -- Warning
                    \t3 -- Error"""
    )

CAFFE_PYTHON = os.environ.get("CAFFE_PYTHON")
if CAFFE_PYTHON:
    sys.path.append(os.path.abspath(CAFFE_PYTHON))
if importlib.util.find_spec("caffe"):
    import caffe
elif CAFFE_PYTHON:
    raise ImportError(
        "Cannot find SSD py-caffe in '{}'. Make sure py-caffe is properly compiled there.".format(
            CAFFE_PYTHON
        )
    )
else:
    raise ImportError(
        "Install py-caffe, set PYTHONPATH to point to py-caffe, or set enviroment variable CAFFE_PYTHON."
    )

from caffe.proto import caffe_pb2


LOGOEXCLUDE = ["Garbage", "Messy", "MessyDark"]


class CaffeClassifier(ImagesModule):
    def __init__(
        self,
        server_name,
        version,
        net_data_dir,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None,
        confidence_min=0.1,
        confidence_min_dict=None,
        layer_name="prob",
        top_n=1,
        postprocess_predictions=None,
        postprocess_args=None,
        gpuid=0,
    ):

        super().__init__(
            server_name,
            version,
            prop_type=prop_type,
            prop_id_map=prop_id_map,
            module_id_map=module_id_map,
        )

        self.confidence_min = confidence_min
        if confidence_min_dict is None:
            confidence_min_dict = {}
        self.layer_name = layer_name
        self.top_n = top_n
        self.postprocess_override = postprocess_predictions
        self.postprocess_args = postprocess_args
        if not isinstance(self.postprocess_args, tuple):
            self.postprocess_args = ()

        if not self.prop_type:
            self.prop_type = "label"

        log.info("Constructing CaffeClassifier")
        caffe.set_mode_gpu()
        caffe.set_device(gpuid)

        labels_file = os.path.join(net_data_dir, "labels.txt")
        try:
            with open(labels_file) as f:
                self.labels = f.read().strip().splitlines()
        except Exception as err:
            log.error("Unable to parse file: " + labels_file)
            log.error(traceback.format_exc())
            raise ValueError(err)

        self.labels = {idx: label for idx, label in enumerate(self.labels)}
        # Set min conf for all labels to 0, but exclude logos in LOGOEXCLDUE
        self.min_conf_filter = {}
        for idx, label in self.labels.items():
            min_conf = self.confidence_min
            if isinstance(confidence_min_dict.get(label), numbers.Number):
                min_conf = confidence_min_dict[label]
            self.min_conf_filter[label] = min_conf

        self.net = caffe.Net(
            os.path.join(net_data_dir, "deploy.prototxt"),
            os.path.join(net_data_dir, "model.caffemodel"),
            caffe.TEST,
        )
        mean_file = os.path.join(net_data_dir, "mean.binaryproto")
        self.transformer = caffe.io.Transformer(
            {"data": self.net.blobs["data"].data.shape}
        )
        if os.path.exists(mean_file):
            blob_meanfile = caffe.proto.caffe_pb2.BlobProto()
            data_meanfile = open(mean_file, "rb").read()
            blob_meanfile.ParseFromString(data_meanfile)
            meanfile = np.squeeze(np.array(caffe.io.blobproto_to_array(blob_meanfile)))
            self.transformer.set_mean("data", meanfile)
        self.transformer.set_transpose("data", (2, 0, 1))

    def process_images(self, images, tstamps, prev_regions):
        # log.debug("Processing images")
        # log.debug("tstamps: "  + str(tstamps))
        assert len(images) == len(tstamps) == len(prev_regions)
        for i, (frame, tstamp, prev_region) in enumerate(
            zip(images, tstamps, prev_regions)
        ):
            log.debug("caffe classifier tstamp: " + str(tstamp))
            try:
                if prev_region is not None:
                    frame = crop_image_from_bbox_contour(frame, prev_region.get("contour"))

                im = self.transformer.preprocess("data", frame)
                self.net.blobs["data"].data[...] = im

                # TODO : clean this up
                probs = self.net.forward()[self.layer_name]
                # log.debug("probs: " + str(probs))
                # log.debug("probs.shape: " + str(probs.shape))
                target_shape = (1, len(self.labels))
                if (probs.shape == target_shape) is False:
                    log.debug(
                        "Changing shape " + str(probs.shape) + "->" + str(target_shape)
                    )
                    probs = np.reshape(probs, target_shape)

                props = []
                for p in probs:
                    # log.debug('p: ' + str(p))
                    p_indexes = np.argsort(p)
                    p_indexes = np.flip(p_indexes, 0)
                    while True:
                        if len(p_indexes) == 1:
                            break
                        index = p_indexes[0]
                        label = self.labels[index]
                        log.debug("label: " + str(label))
                        if label in LOGOEXCLUDE:
                            p_indexes = np.delete(p_indexes, 0)
                        else:
                            break
                    p_indexes = p_indexes[:self.top_n]

                    # log.debug("p_indexes: " + str(p_indexes))

                    for i, property_id in enumerate(p_indexes):
                        if i == self.top_n:
                            break
                        index = p_indexes[i]
                        label = self.labels[index]
                        confidence = p[index]

                        # TODO remove this unknown

                        if confidence < self.confidence_min:
                            label = "Unknown"
                        prop = Property(
                            server=self.name,
                            ver=self.version,
                            value=label,
                            property_type=self.prop_type,
                            confidence=float(confidence),
                            confidence_min=float(self.confidence_min),
                        )
                        if prev_region is not None:
                            prev_region.get("props").append(prop)
                        else:
                            props.append(prop)
                if prev_region is None:
                    self.response.append_region(
                        t=tstamp, region=Region(props=props)
                    )
            except Exception as e:
                log.error(traceback.print_exc())
                log.error(e)



"""MMDetection detector
https://github.com/open-mmlab/mmdetection/
"""
import os
import sys
import glog as log
import numpy as np
import traceback

glog_level = os.environ.get("GLOG_minloglevel", None)

if glog_level is None:
    os.environ["GLOG_minloglevel"] = "1"
    log.info("GLOG_minloglevel isn't set. Setting level to 1 (info)")
    log.info(
        "\nGLOG_minloglevel levels are...\n\
                0 -- Debug\n\
                1 -- Info\n\
                2 -- Warning\n\
                3 -- Error"
    )

try:
    import mmcv
except ImportError:
    raise ImportError("package: mmcv not found")

try:
    from mmdet.apis import init_detector, inference_detector
except ImportError:
    raise ImportError("module: mmdet.apis not found")
    
from multivitamin.module import ImagesModule
from multivitamin.data.response.utils import (
    crop_image_from_bbox_contour,
    compute_box_area
)
from multivitamin.data.response.dtypes import (
    create_bbox_contour_from_points,
    Region,
    Property,
)
from multivitamin.applications.utils import load_idmap


class MMDetector(ImagesModule):
    def __init__(
        self,
        server_name,
        version,
        config_file,
        model_file,
        confidence_min=0.3,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None,
        gpuid=0,
    ):
        """Inference module for https://github.com/open-mmlab/mmdetection/
        
        Args:
            server_name (str): module name
            version (str): version
            config_file (str): path to config_file
            model_file (str): path to model_file
            confidence_min (float, optional): Defaults to 0.3.
            prop_type (str, optional): Defaults to None.
            prop_id_map (dict, optional): Defaults to None.
            module_id_map (dict, optional): Defaults to None.
            gpuid (int, optional): Defaults to 0.
        """
        super().__init__(
            server_name,
            version,
            prop_type=prop_type,
            prop_id_map=prop_id_map,
            module_id_map=module_id_map,
        )
        self.confidence_min = confidence_min
        if not self.prop_type:
            self.prop_type = "object"

        log.info(f"Loading config_file: {config_file} and model_file: {model_file}")
        cfge = mmcv.Config.fromfile(config_file)
        log.debug(f"config: {cfge}")
        self.model = None
        if not (os.path.exists(config_file) and os.path.exists(model_file)):
            raise ValueError("config_file and/or model_file does not exist")
        try:
            self.model = init_detector(config_file, model_file, device="cuda:{}".format(gpuid))
            dummy_image= np.zeros((720, 1280, 3), dtype = "uint8")
            images=[dummy_image]
            tstamps=[0.0]
            self.process_images(images, tstamps)
            log.info('dummy forward pass successful')
        except:
            raise ValueError("Could not init_detector")

    def process_images(self, images, tstamps, prev_regions=None):
        for frame, tstamp in zip(images, tstamps):
            predictions = inference_detector(self.model, frame)

            for pred_idx, pred in enumerate(predictions):
                try:
                    for xmin, ymin, xmax, ymax, confidence in pred:
                        if confidence < self.confidence_min:
                            continue
                        label = self.model.CLASSES[pred_idx]

                        xmin_n = xmin / frame.shape[1]
                        xmax_n = xmax / frame.shape[1]
                        ymin_n = ymin / frame.shape[0]
                        ymax_n = ymax / frame.shape[0]

                        contour = create_bbox_contour_from_points(
                            xmin_n, ymin_n, xmax_n, ymax_n, bound=True
                        )

                        area = compute_box_area(contour)

                        prop = Property(
                                confidence=float(confidence),
                                confidence_min=self.confidence_min,
                                ver=self.version,
                                server=self.name,
                                value=label,
                                property_type=self.prop_type,
                                fraction=area,
                            )

                        self.response.append_region(
                            t=tstamp, region=Region(contour=contour, props=[prop])
                        )

                except Exception:
                    log.error(traceback.format_exc())



# import glog as log
# import inspect
# media_retriever_deprecation = [
#     "from multivitamin.media.media_api import MediaRetriever"
#     ]
# if inspect.stack()[-1].code_context in media_retriever_deprecation:
#     log.warning("Importing MediaRetriever will be deprecated. Please import "
#                 "get_media_retriever from multivitamin.media")

from .opencv_media_retriever import OpenCVMediaRetriever
from .pims_media_retriever import PIMSMediaRetriever
from .file_retriever import FileRetriever
from .opencv_media_retriever import OpenCVMediaRetriever as MediaRetriever



import os
import sys
import argparse

import glog as log

from multivitamin.data.response.frame_drawer import FrameDrawer


if __name__ == "__main__":
    a = argparse.ArgumentParser("python3 visualize_detections --avro test.avro")
    a.add_argument("--avro", type=str, help="path to avro json doc")
    a.add_argument("--dump_dir", default="./tmp")
    # a.add_argument("--dump_video", action="store_true", help="dumps a video file.")
    # a.add_argument("--dump_images", action="store_true", help="dumps image files.")

    args = a.parse_args()
    avro_api = AvroAPI(doc=args.avro)
    fd = FrameDrawer(avro_api=avro_api, pushing_folder=args.dump_dir)
    fd.process(dump_video=False, dump_images=True)



import cv2
import sys

from .media_retriever import AbstractMediaRetriever, AbstractFramesIterator
from .media_retriever import FRAME_EPS


class OpenCVMediaRetriever(AbstractMediaRetriever):
    """A fileretriever for visual media."""

    def __init__(self, url=None):
        """Init MediaRetriever.

        Args:
            url (str | optional): The local or remote url to a file

        """
        super(OpenCVMediaRetriever, self).__init__(url=url)

    @property
    def is_video(self):
        """Check if url is a video."""
        if self.content_type is None:
            return False
        return "video" in self.content_type

    @property
    def is_image(self):
        """Check if url is an image."""
        if self.content_type is None:
            return False
        return "image" in self.content_type and "gif" not in self.content_type

    def _create_video_capture(self):
        """Get the used video capture."""
        if self.is_video:
            self._cap = cv2.VideoCapture(self.url, cv2.CAP_FFMPEG)

        return self._cap

    def _get_fps_from_video_capture(self):
        """Get the fps of the image/video."""
        return self.video_capture.get(cv2.CAP_PROP_FPS)

    def _get_num_frames(self):
        """Get the total number of frames."""
        return self.video_capture.get(cv2.CAP_PROP_FRAME_COUNT)

    def _get_video_frame_shape(self):
        """Get height by width by channels for frames."""
        if not self.video_capture.isOpened():
            self.video_capture.set(cv2.CAP_PROP_POS_FRAMES, 0)

        f = self.get_frame()
        return f.shape

    def _get_frame_from_video(self, tstamp):
        """Return frame if image, or get frame with timestamp in seconds (w/ demicals).

        Note: iterating using get_frames_iterator is significantly faster

        """
        self.video_capture.set(cv2.CAP_PROP_POS_MSEC, tstamp * 1000)
        ret, frame = self.video_capture.read()
        return ret, frame

    def _get_frames_iterator_class(self):
        return OpenCVFramesIterator


class OpenCVFramesIterator(AbstractFramesIterator):
    """Frames iterator object for videos.

    Usage:
    for frame, tstamp in frames_iter:
        do_something(frame)

    """

    def __init__(self,
                 video_cap,
                 video_fps,
                 sample_rate=100.0,
                 start_tstamp=0.0,
                 end_tstamp=sys.maxsize):
        """Frames iterator constructor.

        Args:
            video_cap (cv2.VideoCapture): video capture obj
            sample_rate (float): rate to sample video
            start_tstamp (float): start time for iterating
            end_tstamp (float): end time condition for iterating

        """
        super(OpenCVFramesIterator, self).__init__(video_cap,
                                                   video_fps=video_fps,
                                                   sample_rate=sample_rate,
                                                   start_tstamp=start_tstamp,
                                                   end_tstamp=end_tstamp)

    def _move_cursor_to_tstamp(self, tstamp):
        """Iterate over frames."""
        self.cap.set(cv2.CAP_PROP_POS_MSEC, tstamp * 1000)

    def _get_next_frame(self):
        ret = True
        while ret and self.cur_tstamp <= self.end_tstamp:
            tstamp = self.cap.get(cv2.CAP_PROP_POS_MSEC) / 1000.0
            ret = self.cap.grab()
            if (tstamp - self.cur_tstamp + FRAME_EPS) >= (
                self.period
            ) or self.first_frame:
                ret, frame = self.cap.retrieve()
                self.cur_tstamp = tstamp
                self.first_frame = False
                return ret, frame, self._round_tstamp(tstamp)
        return ret, None, None



from setuptools import setup, find_packages

reqs = None
with open("requirements.txt") as rf:
    reqs = rf.readlines()

setup(
    packages=find_packages(
        exclude=["docs*", "tests*", "examples*", "docker*", "tools*"]
    ),
    setup_requires=["pbr", "setuptools", "setuptools_scm"],
    pbr=True,
    use_scm_version=True
)



import json
import os
from io import BytesIO

import glog as log
import numpy as np
import boto3
import cv2

from multivitamin.applications.classifiers import CaffeClassifier
from multivitamin.data.request import Request
from multivitamin.data.response import Response
from utils import generate_fileobj_from_s3_folder

# from multivitamin.module_api.utils import load_idmap


S3_BUCKET_NAME = "vitamin-cv-test-data"
S3_IMAGES_FOLDER = "data/media/images/generic/"
S3_EXPECTED_PREV_RESPONSES = "data/previous_responses"
S3_PREV_DETECTIONS_FOLDER = "data/sample_detections/random_boxes/"
S3_NET_DATA_FOLDER = "models/caffe/bvlc/SportsNHLLogoClassifier-v1.2/"  # DO NOT TOUCH

LOCAL_NET_DATA_DIR = "/tmp/test_caffe_classifier/net_data"


def test_init():
    global cc

    server_name = "TestClassifier"
    version = "0.0.0"
    # net_data_dir = "/tmp/test_caffe_classifier/net_data"
    # prop_type = "label"
    # prop_id_map = {}
    # module_id_map
    if os.path.exists(LOCAL_NET_DATA_DIR):
        try:
            cc = CaffeClassifier(server_name, version, LOCAL_NET_DATA_DIR)
            return
        except Exception as e:
            log.error(e)
            pass
    else:
        os.makedirs(LOCAL_NET_DATA_DIR)

    for key, net_data_bytes in generate_fileobj_from_s3_folder(
        S3_BUCKET_NAME, S3_NET_DATA_FOLDER
    ):
        filename = os.path.basename(key)
        print("{}/{}".format(LOCAL_NET_DATA_DIR, filename))
        with open("{}/{}".format(LOCAL_NET_DATA_DIR, filename), "wb") as file:
            file.write(net_data_bytes.getvalue())

    cc = CaffeClassifier(server_name, version, LOCAL_NET_DATA_DIR)


messages = [
    {
        "url": "https://s3.amazonaws.com/vitamin-cv-test-data/data/media/videos/generic-short/kitti-clip.mp4"
    },
    {
        "url": "https://s3.amazonaws.com/vitamin-cv-test-data/data/media/images/nhl_image.jpg"
    },
]


def test_process():
    for message in messages:
        request = Request(message)
        response = Response(request=request)
        response = cc.process(response)
        log.info(json.dumps(response.dict, indent=2))


# def download_expected_response(path):
#     s3 = boto3.client("s3")
#     filelike = BytesIO()
#     s3.download_fileobj(S3_BUCKET_NAME, S3_EXPECTED_PREV_RESPONSES+"/"+path, filelike)
#     filelike.seek(0)
#     expected_json = json.loads(filelike.read().decode())
#     return expected_json

# def test_consistency1():
#     # Just an image
#     pass

# def test_consistency2():
#     #  Just a video
#     expected_json = download_expected_response("classification_doc_without_prev_resp.json")

#     message = {
#         "url":expected_json["media_annotation"]["url"]
#     }
#     module_map ={}
#     module_map['NHLPlacementDetector'] =60
#     module_map['NHLLogoClassifier']=61
#     module_map['NBAPlacementDetector'] =69
#     module_map['NBALogoClassifier']=70
#     module_map['NBALeagueDetector']=71

#     sponsor_id_map_file= LOCAL_NET_DATA_DIR+"/idmap.txt"
#     sponsor_map=load_idmap(sponsor_id_map_file)

#     cc = CaffeClassifier("NHLLogoClassifier", "0.0.2", LOCAL_NET_DATA_DIR,prop_type="logo",prop_id_map=sponsor_map,module_id_map=module_map)
#     cc.process(message)
#     cc.update_response()
#     j1 = json.dumps(expected_json, indent=2, sort_keys=True)
#     j2 = json.dumps(cc.avro_api.doc, indent=2, sort_keys=True)
#     j2 = j2.replace(cc.avro_api.doc["media_annotation"]["codes"][0]["date"], expected_json["media_annotation"]["codes"][0]["date"])
#     j2 = j2.replace(cc.avro_api.doc["media_annotation"]["codes"][0]["id"], expected_json["media_annotation"]["codes"][0]["id"])
#     assert(j1 == j2)
#     assert(json.loads(j1) == json.loads(j2))

# def test_consistency3():
#     # A video with previous response
#     expected_json = download_expected_response("classification_doc_with_prev_resp.json")
#     prev_resp = download_expected_response("detection_doc.json")

#     message = {
#         "url":expected_json["media_annotation"]["url"],
#         "prev_response":json.dumps(prev_resp),
#         "bin_decoding":False
#     }
#     module_map ={}
#     module_map['NHLPlacementDetector'] =60
#     module_map['NHLLogoClassifier']=61
#     module_map['NBAPlacementDetector'] =69
#     module_map['NBALogoClassifier']=70
#     module_map['NBALeagueDetector']=71

#     sponsor_id_map_file= LOCAL_NET_DATA_DIR+"/idmap.txt"
#     sponsor_map=load_idmap(sponsor_id_map_file)

#     p={}
#     p['server']="NHLPlacementDetector"
#     p['property_type']='placement'
#     prev_pois=[p]

#     cc = CaffeClassifier("NHLLogoClassifier", "0.0.2", LOCAL_NET_DATA_DIR,prop_type="logo",prop_id_map=sponsor_map,module_id_map=module_map)
#     cc.set_prev_pois(prev_pois=prev_pois)
#     cc.process(message)
#     cc.update_response()
#     j1 = json.dumps(expected_json, indent=2, sort_keys=True)
#     j2 = json.dumps(cc.avro_api.doc, indent=2, sort_keys=True)
#     #print(j2)
#     j2 = j2.replace(cc.avro_api.doc["media_annotation"]["codes"][1]["date"], expected_json["media_annotation"]["codes"][1]["date"])
#     j2 = j2.replace(cc.avro_api.doc["media_annotation"]["codes"][1]["id"], expected_json["media_annotation"]["codes"][1]["id"])
#     assert(j1 == j2)
#     assert(json.loads(j1) == json.loads(j2))



"""Test a 2 module sequence on a short video of cars

1) object detector
2) make-model classifier after filtering only for car detections


TEST WHERE THERE IS NO PREV PROPERTIES OF INTEREST
"""

import os
import sys
import json

import glog as log
import requests
import boto3
import zipfile

from multivitamin.applications.detectors import SSDDetector
from multivitamin.applications.classifiers import CaffeClassifier
from multivitamin.data.request import Request
from multivitamin.data.response import Response
from multivitamin.data.response.io import load_response_from_request
from utils import generate_fileobj_from_s3_folder

S3_BUCKET_NAME = "vitamin-cv-test-data"
CAFFE_S3_NET_DATA_FOLDER = "models/caffe/SportsNHLLogoClassifier-v1.2/"
SSD_S3_NET_DATA_FOLDER = "models/caffe/ssd/SportsNHLPlacementDetector-v1.2/"


def load():
    ssd = None
    ssd_server_name = "TestSSDClassifier"
    ssd_version = "0.0.0"
    ssd_net_data_dir = "/tmp/test_ssd_detector/net_data"
    if not os.path.exists(ssd_net_data_dir):
        os.makedirs(ssd_net_data_dir)

        for key, net_data_bytes in generate_fileobj_from_s3_folder(
            S3_BUCKET_NAME, SSD_S3_NET_DATA_FOLDER
        ):
            filename = os.path.basename(key)
            log.info("{}/{}".format(ssd_net_data_dir, filename))
            with open("{}/{}".format(ssd_net_data_dir, filename), "wb") as file:
                file.write(net_data_bytes.getvalue())

    log.info("Loading SSD")
    ssd = SSDDetector(ssd_version, ssd_version, ssd_net_data_dir)

    cc = None
    server_name = "TestClassifier"
    version = "0.0.0"
    net_data_dir = "/tmp/test_caffe_classifier/net_data"
    # prop_type = "label"
    # prop_id_map = {}
    # module_id_map
    if not os.path.exists(net_data_dir):
        os.makedirs(net_data_dir)

        for key, net_data_bytes in generate_fileobj_from_s3_folder(
            S3_BUCKET_NAME, CAFFE_S3_NET_DATA_FOLDER
        ):
            filename = os.path.basename(key)
            log.info("{}/{}".format(net_data_dir, filename))
            with open("{}/{}".format(net_data_dir, filename), "wb") as file:
                file.write(net_data_bytes.getvalue())

    log.info("Loading Caffe")
    cc = CaffeClassifier(server_name, version, net_data_dir)

    return ssd, cc


def test_process():
    # Test on short video
    ssd, cc = load()
    message = {
        "url": "https://s3.amazonaws.com/video-ann-testing/short_flamesatblues.mp4",
        "bin_encoding": "false",
        "bin_decoding": "false",
    }
    request = Request(message)
    response = load_response_from_request(request)
    log.info("SSD DETECTOR")
    response = ssd.process(request, response)
    log.info("CAFFE CLASSIFIER")
    cc.set_prev_props_of_interest(
        [
            {
                "property_type": "placement",
                "company": "gumgum",
                "server": "TestSSDClassifier",
                "value": "Static Dasherboard",
            }
        ]
    )
    response = cc.process(request, response)

    doc = response.dict
    log.info(f"doc: {json.dumps(doc, indent=2)}")
    with open("t.json", "w") as wf:
        json.dump(doc, wf)


test_video = "https://s3.amazonaws.com/video-ann-testing/kitti-clip.mp4"
dst_video = os.path.join("/tmp", os.path.basename(test_video))
# log.setLevel("DEBUG")


def test_app():
    _download_video()
    _download_obj_det()
    _download_clf()
    req = {"url": test_video, "sample_rate": 5.0}
    req = RequestAPI(req)
    log.info("Processing request: {}".format(req))
    ssd = SSDDetector(
        server_name="ObjectDetector", version="1.0", net_data_dir="/tmp/ssd/net_data/"
    )
    resp = ssd.process(req)
    assert resp == "SUCCESS"
    ssd.update_response()

    req.reset_media_api()
    clf = CaffeClassifier(
        server_name="MakeModelClassifier",
        version="1.0",
        net_data_dir="/tmp/clf/net_data",
    )
    clf.set_prev_pois({"value": "car"})
    resp = clf.process(req)
    assert resp == "SUCCESS"
    out = req.get_avro_api().get_response()
    print(json.dumps(out, indent=2))


def _download_video():
    log.info("Downloading video")
    r = requests.get(test_video)
    log.info("status code: {}".format(r.status_code))
    assert r.status_code == 200
    with open(dst_video, "wb") as f:
        f.write(r.content)
    assert os.path.exists(dst_video)
    log.info("{}".format(dst_video))


def _download_obj_det():
    log.info("Downloading object detector model")
    s3_client = boto3.client("s3")
    tmp_folder = "/tmp/ssd"
    dl_file = "{}/net_data.zip".format(tmp_folder)
    if not os.path.exists(tmp_folder):
        os.makedirs(tmp_folder)
    s3_client.download_file(
        "cvapis-data", "ssd-detector/objectdetector/net_data.zip", dl_file
    )
    log.info("Model downloaded")
    with open(dl_file, "rb") as f:
        log.info("Unzipping")
        z = zipfile.ZipFile(f)
        for name in z.namelist():
            z.extract(name, "/tmp/ssd/")
    log.info("Unzipping... Done")


def _download_clf():
    log.info("Downloading classification model")
    s3_client = boto3.client("s3")
    tmp_folder = "/tmp/clf"
    dl_file = "{}/net_data.zip".format(tmp_folder)
    if not os.path.exists(tmp_folder):
        os.makedirs(tmp_folder)
    s3_client.download_file(
        "cvapis-data", "classifiers/makemodelclassifier/net_data.zip", dl_file
    )
    log.info("Model downloaded")
    with open(dl_file, "rb") as f:
        log.info("Unzipping")
        z = zipfile.ZipFile(f)
        for name in z.namelist():
            z.extract(name, "/tmp/clf/")
    log.info("Unzipping... Done")



import glog as log
import boto3
import json

from multivitamin.data import Request
from multivitamin.apis.comm_api import CommAPI
from multivitamin.apis import config

boto3.client("iam")  # required to expose iam roles to boto3?


class SQSAPI(CommAPI):
    def __init__(self, queue_name):
        """Class to pull messages from and push messages to SQS 

        Args:
            queue_name (str): SQS name in AI account
        """
        assert(isinstance(queue_name, str))
        self.sqs = boto3.client("sqs")
        log.info(f"Attempting to retrieve queue: {queue_name}")
        try:
            queue = self.sqs.get_queue_url(QueueName=queue_name)
            log.debug(f"Retrieved queue: {queue}")
        except Exception:
            log.error("The queue does not exist, creating it")
            queue = self.sqs.create_queue(
                QueueName=queue_name, Attributes={"DelaySeconds": "120"}
            )
        self.queue_url = queue["QueueUrl"]
        log.info(f"Retrieved queue_url: {self.queue_url}")

    def pull(self, n=1):
        """Pull messages from SQS queue

        Args:
            n (int): batch pulling. NOT IMPLEMENTED
        
        Returns:
            list[Request]: list of requests
        """
        log.info(f"Polling request from queue {self.queue_url}...")
        response = self.sqs.receive_message(
            QueueUrl=self.queue_url, WaitTimeSeconds=config.SQS_WAIT_TIME_SEC
        )
        while "Messages" not in response:
            log.debug(f"Polling request from queue {self.queue_url}...")
            response = self.sqs.receive_message(
                QueueUrl=self.queue_url, WaitTimeSeconds=config.SQS_WAIT_TIME_SEC
            )
        log.debug(f"sqs.receive_message response: {response}")
        requests = []
        for m in response["Messages"]:
            log.debug(str(m))
            requests.append(
                Request(request_input=m["Body"], request_id=m["ReceiptHandle"])
            )
        return requests

    def push(self, request):
        raise NotImplementedError("Intentionally not implemented--we don't want to push resposnes to SQS")

    def delete_message(self, request_id):
        """Delete a message from the SQS queue given a request_id

        Args:
            request_id: request_id from sqs.receive_message
        """
        self.sqs.delete_message(QueueUrl=self.queue_url, ReceiptHandle=request_id)

    @staticmethod       
    def delete_message_thead_safe(request_id,queue_url):
        """Delete a message from the SQS queue given a request_id

        Args:
            request_id: request_id from sqs.receive_message
            queueurl: the url of the queue from which to delete the message
        """
        sqs = boto3.client("sqs")
        sqs.delete_message(QueueUrl=queue_url, ReceiptHandle=request_id)


import os
import requests
import magic
from io import BytesIO
import urllib.parse
import glog as log
from imohash import hashfileobject
from .http_fileobj import HTTPFile


class FileRetriever:
    """A generic class for retrieving files."""

    def __init__(self, url=None):
        """Init FileRetriever.

        Args:
            url (str | optional): A local or remote url to a file.

        """
        self._url = None
        self._is_local = None
        self._content_type = None
        self._hash = None
        if url is not None:
            self.url = url

    @property
    def url(self):
        """Get the url to the file."""
        return self._url

    @url.setter
    def url(self, value):
        self._content_type = None
        url_scheme = urllib.parse.urlparse(value).scheme

        self._url = value
        if url_scheme in ["", "file"]:
            self._is_local = True
        else:
            self._is_local = False

        if not self.exists:
            self._is_local = None
            self._url = None
            raise FileNotFoundError(value)

    @property
    def exists(self):
        """Check if the url points to a real file."""
        if self.is_local:
            return self._does_local_file_exist()
        else:
            return self._does_remote_file_exist()

    def _does_local_file_exist(self):
        if os.path.isfile(self.filepath):
            return True
        return False

    def _does_remote_file_exist(self):
        try:
            requests.head(self.url)
        except Exception as e:
            log.warning("Failed to retrieve url: {}".format(e))
            return False
        return True

    @property
    def filepath(self):
        """Get the local filepath.

        Returns:
            `None` if the file is remote
            else will return the local filepath (not the local file url)

        """
        return self.url.replace("file://", "") if self.is_local else None

    @property
    def content_type(self):
        """Get the MIME type of the file."""
        if self._content_type is None and self.is_local and self.exists:
            mime = magic.Magic(mime=True)
            self._content_type = mime.from_file(self.filepath)

        if self._content_type is None and self.is_remote and self.exists:
            resp = requests.head(self.url)
            self._content_type = resp.headers["Content-Type"]

        return self._content_type

    @property
    def is_local(self):
        """Check if file is local."""
        return self._is_local

    @property
    def is_remote(self):
        """Check if file is remote."""
        return not self.is_local if isinstance(self.is_local, bool) else False

    @property
    def filename(self):
        """Get the filename of the url."""
        return os.path.basename(self.url)

    @property
    def hash(self):
        """Get quick hash of file bytes."""
        if self._hash is None:
            filelike = HTTPFile(self.url)
            self._hash = hashfileobject(filelike, hexdigest=True)

        return self._hash

    def download(self, filepath=None, return_filelike=False):
        """Download file to filepath.

        Args:
            filepath (str | optional): Filepath to write file to.
                                If directory, it will take it's original
                                filename and save to that directory
            filelike (bool | optional): To return a filelike object or not

        Returns:
            filelike_obj: A BytesIO object containing the file bytes
                            (only if return_filelike is True)
        """

        if self.is_remote:
            response = requests.get(self.url)
            filelike_obj = BytesIO(response.content)
        else:
            with open(self.filepath, "rb") as f:
                filelike_obj = BytesIO(f.read())

        path = None
        if isinstance(filepath, str):
            if os.path.isdir(filepath):
                path = "{}/{}".format(filepath, self.filename)
            else:
                path = filepath
            with open(path, "wb") as f:
                f.write(filelike_obj.read())

        if return_filelike is True:
            filelike_obj.seek(0)
            return filelike_obj



import sys
import numpy as np
from PIL import Image
import glog as log
import math
from abc import ABC, abstractmethod
from .file_retriever import FileRetriever

FRAME_EPS = 0.001
DECIMAL_SIGFIG = 3


class AbstractMediaRetriever(FileRetriever, ABC):
    """A fileretriever for visual media."""

    def __init__(self, url=None):
        """Init MediaRetriever.

        Args:
            url (str | optional): The local or remote url to a file.

        """
        self._cap = None
        self._image = None
        self._shape = None

        super(AbstractMediaRetriever, self).__init__(url=url)

    @FileRetriever.url.setter
    def url(self, value):
        """Set the image/video url."""
        FileRetriever.url.fset(self, value)
        self._cap = None
        self._image = None
        self._shape = None
        if not self.is_image and not self.is_video:
            raise ValueError(
                "Unsupported Content-Type: {}\n Expected one of these options:\n {}, {}".format(
                    self.content_type, "image/*", "video/*"
                )
            )

        w, h = self.get_w_h()
        if w in [0, None] or h in [0, None]:
            raise ValueError(
                "Unable to load visial media properly: {}".format(self.url)
            )

    @property
    @abstractmethod
    def is_video(self):
        """Check if url is a video."""
        pass

    @property
    @abstractmethod
    def is_image(self):
        """Check if url is an image."""
        pass

    @property
    def video_capture(self):
        """Get object that has all of visual information."""
        if self._cap is None:
            self._create_video_capture()

        return self._cap

    @abstractmethod
    def _create_video_capture(self):
        pass

    @property
    def fps(self):
        """Get the fps of the image/video."""
        if self.is_image:
            log.error("Images have no fps")
            return None

        if not self.video_capture:
            log.error("VideoCapture is None")
            return None

        return self._get_fps_from_video_capture()

    @abstractmethod
    def _get_fps_from_video_capture(self):
        pass

    @property
    def total_frames(self):
        """Get the total number of frames."""
        if self.is_image:
            return 1

        if not self.video_capture:
            log.error("VideoCapture is None")
            return None

        return self._get_num_frames()

    @abstractmethod
    def _get_num_frames(self):
        pass

    @property
    def length(self):
        """Get the temporal length of the image/video."""
        if self.is_image:
            return 0

        if self.fps == 0:
            return 0

        return float(self.total_frames) / self.fps

    @property
    def shape(self):
        """Get height by width by channels for frames."""
        if self._shape is None:
            if self.is_image:
                self._shape = self.image.shape

            if self.is_video:
                self._shape = self._get_video_frame_shape()

        if self._shape is None:
            return (None, None, None)

        return self._shape

    @abstractmethod
    def _get_video_frame_shape(self):
        pass

    @property
    def image(self):
        """Get the image.

        Return `None` if it's a video.

        """
        if not self.is_image:
            return None

        if self._image is not None:
            return self._image

        filelike_obj = self.download(return_filelike=True)
        image = np.array(Image.open(filelike_obj).convert('RGB'))
        if image.shape[2] > 3:
            log.warning("Image has >3 channels. Cropping to 3.")
            image = image[:, :, :3]
        self._image = image[:, :, ::-1].copy()
        return self._image

    def tstamp_to_frame_index(self, tstamp):
        """Convert a timestamp to a frame index.

        Args:
            tstamp (float): A float in seconds.

        Returns:
            int: The nearest frame to that tstamp.

        """
        return round(tstamp * self.fps)

    def get_frame(self, tstamp=0.0):
        """Return frame if image, or get frame with timestamp in seconds (w/ demicals).

        Note: iterating using get_frames_iterator is significantly faster

        """
        if not self.url:
            raise ValueError('URL not set. Please use med_ret.set_url("...")')

        if self.is_image:
            return self.image

        elif self.is_video:
            ret, frame = self._get_frame_from_video(tstamp=tstamp)
            if not ret:
                return ret
            return frame

    @abstractmethod
    def _get_frame_from_video(self, tstamp):
        pass

    @abstractmethod
    def _get_frames_iterator_class(self):
        pass

    def get_frames_iterator(
        self, sample_rate=100.0, start_tstamp=0.0, end_tstamp=sys.maxsize
    ):
        """Get a frames iterator.

        If image, returns a list of length 0 with a tuple (image, tstamp),
        If video, returns a FramesIterator

        Usage in either image or video case:
        for frame, tstamp in med_ret.get_frames_iterator():
            do_something(frame, tstamp)

        Args:
            sample_rate (float): sample rate for extracting frames from video
            start_tstamp (float): starting timestamp for iteration
            end_tstamp (float): ending timestamp for iteration

        Returns:
            iterator

        Raises:
            ValueError: If URL is not set

        """
        if not self.url:
            raise ValueError('URL not set. Please use med_ret.set_url("...")')

        if self.is_image:
            return [(self.image, 0.00)]

        elif self.is_video:
            fi = self._get_frames_iterator_class()
            return fi(self.video_capture,
                      self.fps,
                      sample_rate,
                      start_tstamp,
                      end_tstamp)

    def get_length(self):
        """Get the temporal length of the image/video."""
        return self.length

    def get_num_frames(self):
        """Get the number of frame in image/video."""
        return self.total_frames

    def get_fps(self):
        """Get the fps of the image/video."""
        return self.fps

    def get_w_h(self):
        """Get the width and height the visual media."""
        return self.shape[0:2][::-1]


class AbstractFramesIterator(ABC):
    """Frames iterator object for videos.

    Usage:
    for frame, tstamp in frames_iter:
        do_something(frame)

    Note: Do not use multiple times/attempt to restart from the start of the video. This feature is incomplete.
    TODO: check to see if start_tstamp & end_tstamp checks cause significant slowing

    """

    def __init__(self,
                 video_cap,
                 video_fps,
                 sample_rate=100.0,
                 start_tstamp=0.0,
                 end_tstamp=sys.maxsize):
        """Frames iterator constructor.

        Args:
            video_cap (cv2.VideoCapture): video capture obj
            fps (float): The video fps.
            sample_rate (float): rate to sample video
            start_tstamp (float): start time for iterating
            end_tstamp (float): end time condition for iterating

        """
        self.cap = video_cap
        self.period = max(1.0 / sample_rate, 1.0 / video_fps)
        log.debug("Period: {}".format(self.period))
        self.start_tstamp = start_tstamp
        self.end_tstamp = end_tstamp
        self.cur_tstamp = self.start_tstamp
        self.first_frame = True

        self._move_cursor_to_tstamp(self.start_tstamp)
        log.debug("Setting start_tstamp of iterator to {}".format(start_tstamp))

    @abstractmethod
    def _move_cursor_to_tstamp(self, tstamp):
        pass

    def __iter__(self):
        """Iterate over frames."""
        self.cur_tstamp = self.start_tstamp
        self.first_frame = True
        self._move_cursor_to_tstamp(self.start_tstamp)
        return self

    def _round_tstamp(self, tstamp):
        tstamp = math.floor(tstamp * 100000.0) / 100000.0
        return round(tstamp, DECIMAL_SIGFIG)

    @abstractmethod
    def _get_next_frame(self):
        pass

    def __next__(self):
        """Get next frame."""
        ret = True
        frame = None
        while ret and self.cur_tstamp <= self.end_tstamp:
            ret, frame, tstamp = self._get_next_frame()
            if ret:
                return frame, tstamp

        log.info("No more frames to read")
        raise StopIteration()



import random
import datetime
import glog as log
import json

from multivitamin.data.response import config

# from multivitamin.data.response.dtypes import Point


def read_json(file_path):
    """Convenience method for reading jsons"""
    return json.load(open(file_path))


def write_json(json_str, file_path, indent=2):
    """Convenience method for writing jsons"""
    with open(file_path, "w") as wf:
        if type(json_str) is dict:
            json.dump(json_str, wf, indent=indent)
        elif type(json_str) is str:
            wf.write(json_str)
        else:
            raise ValueError(
                "json_str input is not a str or dict. Of type: {}".format(type(json_str))
            )


def p0p1_from_bbox_contour(contour, w=1, h=1, dtype=int):
    """Convert cv_schema `contour` into p0 and p1 of a bounding box.

    Args:
        contour (list): list dict of points x, y
        w (int): width
        h (int): height

    Returns:
        Two points dict(x, y): p0 (upper left) and p1 (lower right)
    """
    if len(contour) != 4:
        log.error("To use p0p1_from_bbox_contour(), input must be a 4 point bbox contour")
        return None

    # Convert number of pixel to max pixel index
    w_max_px_ind = max(w - 1, 1)
    h_max_px_ind = max(h - 1, 1)

    x0 = float(contour[0]["x"])
    y0 = float(contour[0]["y"])
    x1 = float(contour[0]["x"])
    y1 = float(contour[0]["y"])
    for pt in contour:
        x0 = min(x0, float(pt["x"]))
        y0 = min(y0, float(pt["y"]))
        x1 = max(x1, float(pt["x"]))
        y1 = max(y1, float(pt["y"]))

    x0 = dtype(x0 * w_max_px_ind)
    y0 = dtype(y0 * h_max_px_ind)
    x1 = dtype(x1 * w_max_px_ind)
    y1 = dtype(y1 * h_max_px_ind)
    return (x0, y0), (x1, y1)


def compute_box_area(contour):
    """Function to compute the spatial fraction based on the contour
    
    Args:
        contour (List[Point]): contour
    
    Returns:
        float: fraction
    """
    assert len(contour) == 4
    p0, p1 = p0p1_from_bbox_contour(contour, dtype=float)
    return float((p1[0] - p0[0]) * (p1[1] - p0[1]))


def crop_image_from_bbox_contour(image, contour):
    """Crop an image given a bounding box contour
    
    Args:  
        image (np.array): image
        contour (dict[float]): points of a bounding box countour
    
    Returns:
        np.array: image
    """
    if contour is None:
        return image

    h = image.shape[0]
    w = image.shape[1]
    (x0, y0), (x1, y1) = p0p1_from_bbox_contour(contour, w=w, h=h)
    return image[y0:y1, x0:x1]


def round_float(val):
    """Function to round a float to our set number of sigificant digits

    Args:
        val (float): input value
    
    Returns:
        float: rounded float value
    """
    return round(val, config.SIGFIG)


def round_float_to_str(val):
    """Function to round a float to our set number of sigificant digits

    Args:
        val (float): input value
    
    Returns:
        float: rounded float value
    """
    return "{:.4f}".format(val)


# def round_all_pts_in_contour(contour):
#     """Function to round all pts in a contour

#     Args:
#         contour (list): list of dicts with keys x, y

#     Returns:
#         list: of dicts of points
#     """
#     rounded_contour = []
#     for pt in contour:
#         rounded_contour.append(Point(x=round_float(pt["x"]), y=round_float(pt["y"])))
#     return rounded_contour


def round_all_pts_in_contour_to_str(contour):
    """Function to round all pts in a contour

    Args:
        contour (list): list of dicts with keys x, y
    
    Returns:
        list: of dicts of points
    """
    rounded_contour = []
    for pt in contour:
        pt["x"] = round_float_to_str(pt["x"])
        pt["y"] = round_float_to_str(pt["y"])
        rounded_contour.append(pt)
    return rounded_contour


def points_equal(val0, val1):
    """Function for float quality comparison for point values

    Args:
        val0 (float): first value
        val1 (float): second value
    
    Returns:
        bool: equality
    """
    return abs(round_float(val0) - round_float(val1)) < config.POINT_EPS


def times_equal(val0, val1, eps=config.TIME_EPS):
    """Function for float quality comparison for time values

    Args:
        val0 (float): first value
        val1 (float): second value
    
    Returns:
        bool: equality
    """
    if eps is None:
        eps = config.TIME_EPS
    return abs(round_float(val0) - round_float(val1)) < eps


def get_current_time():
    """Get string YYYYMMDDHHMMSS for current time

    Returns:
        current_time (str)
    """
    now = datetime.datetime.now()
    date = "{:4}{:2}{:2}{:2}{:2}{:2}".format(
        str(now.year).zfill(4),
        str(now.month).zfill(2),
        str(now.day).zfill(2),
        str(now.hour).zfill(2),
        str(now.minute).zfill(2),
        str(now.second).zfill(2),
    )
    return date


def get_current_date():
    """Get string YYYYMMDD for current date

    Returns:
        current_date (str)
    """
    now = datetime.datetime.now()
    date = "{:4}{:2}{:2}".format(
        str(now.year).zfill(4), str(now.month).zfill(2), str(now.day).zfill(2)
    )
    return date


# def jaccard_distance_between_contours(contourA, contourB, w, h):
#     pass


def intersection_between_bboxes(bbox0, bbox1):
    assert len(bbox0) == len(bbox1) == 4
    if type(bbox0) == type([]):
        bbox0 = p0p1_from_bbox_contour(bbox0, dtype=float)
    if type(bbox1) == type([]):
        bbox1 = p0p1_from_bbox_contour(bbox1, dtype=float)
    p00, p01 = bbox0
    p10, p11 = bbox1

    max_x0 = max(p00[0], p10[0])
    max_y0 = max(p00[1], p10[1])
    min_x1 = min(p01[0], p11[0])
    min_y1 = min(p01[1], p11[1])
    return max(max(min_x1 - max_x0, 0) * max(min_y1 - max_y0, 0), 0)


def union_and_intersection_between_bboxes(bbox0, bbox1):
    assert len(bbox0) == len(bbox1) == 4
    if type(bbox0) == type([]):
        bbox0 = p0p1_from_bbox_contour(bbox0, dtype=float)
    if type(bbox1) == type([]):
        bbox1 = p0p1_from_bbox_contour(bbox1, dtype=float)

    p00, p01 = bbox0
    p10, p11 = bbox1
    area0 = (p01[0] - p00[0]) * (p01[1] - p00[1])
    area1 = (p11[0] - p10[0]) * (p11[1] - p10[1])

    intersection = intersection_between_bboxes(bbox0, bbox1)
    union = area0 + area1 - intersection

    return union, intersection


def jaccard_distance_between_bboxes(bbox0, bbox1):
    union, intersection = union_and_intersection_between_bboxes(bbox0, bbox1)
    if union == 0:
        return 0
    return intersection / union



import os
import sys
import glog as log
import numpy as np
import traceback
import inspect
import importlib.util

glog_level = os.environ.get("GLOG_minloglevel", None)

if glog_level is None:
    os.environ["GLOG_minloglevel"] = "1"
    log.info("GLOG_minloglevel isn't set. Setting level to 1 (info)")
    log.info(
        "\nGLOG_minloglevel levels are...\n\
                0 -- Debug\n\
                1 -- Info\n\
                2 -- Warning\n\
                3 -- Error"
    )

if importlib.util.find_spec("tensorflow"):
    import tensorflow as tf
else:
    raise ImportError(
        "Install tensorflowgpu 1.9.0"
    )

from google.protobuf import text_format

from multivitamin.utils.GPUUtilities import GPUUtility
from multivitamin.module import ImagesModule
from multivitamin.data.response.utils import (
    crop_image_from_bbox_contour,
    compute_box_area
)
from multivitamin.data.response.dtypes import (
    create_bbox_contour_from_points,
    Region,
    Property,
    Point
)
from multivitamin.applications.utils import load_idmap, load_label_prototxt

CONFIDENCE_MIN = 0.3


class TFDetector(ImagesModule):
    def __init__(
        self, 
        server_name, 
        version, 
        net_data_dir,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None,
        **gpukwargs
    ):
        super().__init__(
            server_name, 
            version, 
            prop_type=prop_type,
            prop_id_map=prop_id_map,
            module_id_map=module_id_map
        )
        self.server_name = server_name
        self.version = version

        if not self.prop_type:
            self.prop_type="object" 
        gpu_util = GPUUtility(**gpukwargs)
        available_devices = gpu_util.get_gpus()
        log.info("Found GPU devices: {}".format(available_devices))
        if available_devices:
            os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID"
            os.environ["CUDA_VISIBLE_DEVICES"]= ",".join([str(gpu) for gpu in available_devices])
        self.label_map = dict()
        labelmap_file = os.path.join(net_data_dir, 'idmap.txt')
        try:
            with open(labelmap_file, 'r') as file_h:
                for line in file_h:
                    label_id, label_name = line.strip().split('\t')
                    label_name = label_name.strip()
                    if label_id not in self.label_map.keys():
                        self.label_map[label_id] = label_name
        except Exception as e:
            log.error("Unable to parse file: " + labelmap_file)
            log.error(traceback.format_exc())
            exit(1)

        model = os.path.join(net_data_dir,'frozen_inference_graph.pb')
        detection_graph = tf.Graph()
        with detection_graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(model, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')
        self._detection_graph = detection_graph
        cfg = tf.ConfigProto()
        cfg.gpu_options.allow_growth = True
        self._sess =  tf.Session(graph=detection_graph, config=cfg)
        

    def process_images(self, images, tstamps, detections_of_interest=None):
        image_tensor = self._detection_graph.get_tensor_by_name('image_tensor:0')
        detection_boxes = self._detection_graph.get_tensor_by_name('detection_boxes:0')
        detection_scores = self._detection_graph.get_tensor_by_name('detection_scores:0')
        detection_classes = self._detection_graph.get_tensor_by_name('detection_classes:0')
        num_pred_detections = self._detection_graph.get_tensor_by_name('num_detections:0')

        for frame, tstamp in zip(images, tstamps):
            frame_np = np.array(frame)
            frame_np = frame_np[...,::-1]
            frame_np_expanded = np.expand_dims(frame_np, axis=0)
            (detections, scores, classes, num) = self._sess.run([detection_boxes, detection_scores, detection_classes, num_pred_detections], feed_dict={image_tensor: frame_np_expanded})
            num = int(num[0])
            for det_idx in range(num):
                try:
                    confidence = float(scores[0,det_idx])
                    if confidence < CONFIDENCE_MIN:
                        continue
                    xmin = float(detections[0,det_idx,1])
                    ymin = float(detections[0,det_idx,0])
                    xmax = float(detections[0,det_idx,3])
                    ymax = float(detections[0,det_idx,2])
                    contour = [Point(xmin, ymin), Point(xmax, ymin), Point(xmax, ymax), Point(xmin, ymax)]
                    prop = Property(
                           property_type=self.prop_type,
                           server=self.server_name,
                           ver=self.version,
                           value = str(self.label_map[str(int(classes[0,det_idx]))]),
                           confidence=confidence,
                           confidence_min=CONFIDENCE_MIN
                           )
                    region = Region(contour, [prop])
                    self.response.append_region(t=tstamp, region=region)

                except Exception as e:
                    log.error(traceback.format_exc())



"""Single shot multi-box detector in Caffe

Currently, does not support modules in sequence"""
import os
import sys
import glog as log
import numpy as np
import traceback
import inspect
import importlib.util

glog_level = os.environ.get("GLOG_minloglevel", None)

if glog_level is None:
    os.environ["GLOG_minloglevel"] = "1"
    log.info("GLOG_minloglevel isn't set. Setting level to 1 (info)")
    log.info(
        "\nGLOG_minloglevel levels are...\n\
                0 -- Debug\n\
                1 -- Info\n\
                2 -- Warning\n\
                3 -- Error"
    )

SSD_CAFFE_PYTHON = os.environ.get("SSD_CAFFE_PYTHON")
if SSD_CAFFE_PYTHON:
    sys.path.append(os.path.abspath(SSD_CAFFE_PYTHON))

if importlib.util.find_spec("caffe"):
    import caffe
elif SSD_CAFFE_PYTHON:
    raise ImportError(
        f"Cannot find SSD py-caffe in '{SSD_CAFFE_PYTHON}'. Make sure \
          py-caffe is properly compiled there."
    )
else:
    raise ImportError(
        "Install py-caffe, set PYTHONPATH to point to py-caffe, or \
         set enviroment variable SSD_CAFFE_PYTHON."
    )

from google.protobuf import text_format
from caffe.proto import caffe_pb2 as cpb2

from multivitamin.module import ImagesModule
from multivitamin.data.response.utils import (
    crop_image_from_bbox_contour,
    compute_box_area
)
from multivitamin.data.response.dtypes import (
    create_bbox_contour_from_points,
    Region,
    Property,
)
from multivitamin.applications.utils import load_idmap, load_label_prototxt

LAYER_NAME = "detection_out"


class SSDDetector(ImagesModule):
    def __init__(
        self,
        server_name,
        version,
        net_data_dir,
        confidence_min=0.3,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None,
        gpuid=0,
    ):
        super().__init__(
            server_name,
            version,
            prop_type=prop_type,
            prop_id_map=prop_id_map,
            module_id_map=module_id_map,
        )
        self.confidence_min = confidence_min
        if not self.prop_type:
            self.prop_type = "object"

        caffe.set_mode_gpu()
        caffe.set_device(gpuid)

        idmap_file = os.path.join(net_data_dir, "labelmap.prototxt")
        self.labelmap = load_label_prototxt(idmap_file)
        log.info(str(len(self.labelmap.keys())) + " labels parsed.")

        self.net = caffe.Net(
            os.path.join(net_data_dir, "deploy.prototxt"),
            os.path.join(net_data_dir, "model.caffemodel"),
            caffe.TEST,
        )
        self.transformer = caffe.io.Transformer(
            {"data": self.net.blobs["data"].data.shape}
        )

        mean_file = os.path.join(net_data_dir, "mean.binaryproto")
        if os.path.exists(mean_file):
            log.info("Setting meanfile")
            blob_meanfile = caffe.proto.caffe_pb2.BlobProto()
            data_meanfile = open(mean_file, "rb").read()
            blob_meanfile.ParseFromString(data_meanfile)
            meanfile = np.squeeze(np.array(caffe.io.blobproto_to_array(blob_meanfile)))
            self.transformer.set_mean("data", meanfile)
        self.transformer.set_transpose("data", (2, 0, 1))

    def process_images(self, images, tstamps, prev_detections=None):
        for frame, tstamp in zip(images, tstamps):
            im = self.transformer.preprocess("data", frame)
            self.net.blobs["data"].data[...] = im
            predictions = self.net.forward()[LAYER_NAME]

            for pred_idx in range(predictions.shape[2]):
                try:
                    confidence = float(predictions[0, 0, pred_idx, 2])
                    if confidence < self.confidence_min:
                        continue
                    index = int(predictions[0, 0, pred_idx, 1])
                    label = self.labelmap[index]
                    xmin = float(predictions[0, 0, pred_idx, 3])
                    ymin = float(predictions[0, 0, pred_idx, 4])
                    xmax = float(predictions[0, 0, pred_idx, 5])
                    ymax = float(predictions[0, 0, pred_idx, 6])

                    contour = create_bbox_contour_from_points(
                        xmin, ymin, xmax, ymax, bound=True
                    )
                    area = compute_box_area(contour)
                    prop = Property(
                            confidence=confidence,
                            confidence_min=self.confidence_min,
                            ver=self.version,
                            server=self.name,
                            value=label,
                            property_type=self.prop_type,
                            fraction=area,
                        )
                    self.response.append_region(
                        t=tstamp, region=Region(contour=contour, props=[prop])
                    )
                except Exception:
                    log.error(traceback.format_exc())
